﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using OldBvets.TrainDat;
using BvetsTool;

namespace BvetsToolApi
{
    using Bvets5.Vehicle;

    /// <summary>
    /// 算術クラス
    /// </summary>
    /// <remarks>
    /// 計算用のメソッドの集まり
    /// </remarks>
    public static class OrigMath
    {
        /// <summary>
        /// 重力加速度（定数）の取得
        /// </summary>
        public static double G
        {
            get { return 9.80665; }
        }

        /// <summary>
        /// 線形補完
        /// </summary>
        /// <param name="x">変数x</param>
        /// <param name="x1">座標1x</param>
        /// <param name="y1">座標1y</param>
        /// <param name="x2">座標2x</param>
        /// <param name="y2">座標2y</param>
        /// <returns>補完後の座標y</returns>
        public static double LinearInterpolation(double x, double x1, double y1, double x2, double y2)
        {
            return y1 + (x - x1) / (x2 - x1) * (y2 - y1);
        }

        /// <summary>
        /// 非選択グラフのペン色
        /// </summary>
        /// <param name="color">オリジナルのペン色</param>
        /// <returns>変換後ペン色</returns>
        public static Color NonseledtedIndexColor(Color color)
        {
            return Color.FromArgb(color.R / 2, color.G / 2, color.B / 2);
        }

        /// <summary>
        /// 非選択グラフのペン色
        /// </summary>
        /// <param name="color">オリジナルのペン色</param>
        /// <param name="rate">色の明るさの比率</param>
        /// <returns>変換後ペン色</returns>
        public static Color NonseledtedIndexColor(Color color, double rate)
        {
            double r;
            if (rate > 1)
            {
                r = 1;
            }
            else if (rate < 0)
            {
                r = 0;
            }
            else
            {
                r = rate;
            }
            return Color.FromArgb((int)(color.R * r), (int)(color.G * r), (int)(color.B * r));
        }

        /// <summary>
        /// デフォルト走行抵抗係数算出
        /// </summary>
        public static void CalcRunningResistanceFactor()
        {
            // 参照型だと呼び出さなくても連動？
            Parameters.Dynamics.RunningResistanceFactorA = 0.275 + 0.076 * (Parameters.Dynamics.MotorcarCount + Parameters.Dynamics.TrailerCount - 1);
            Parameters.Dynamics.RunningResistanceFactorB = 0.000242 * Parameters.Dynamics.MotorcarCount * Parameters.Dynamics.MotorcarWeight + 0.0000275 * Parameters.Dynamics.TrailerCount * Parameters.Dynamics.TrailerWeight;
            Parameters.Dynamics.RunningResistanceFactorC = 0.0162 * Parameters.Dynamics.MotorcarCount * Parameters.Dynamics.MotorcarWeight + 0.00765 * Parameters.Dynamics.TrailerCount * Parameters.Dynamics.TrailerWeight;
        }

        /// <summary>
        /// ある速度における編成全体に掛かる走行抵抗力[N]
        /// </summary>
        /// <param name="v">速度[N]</param>
        /// <returns>走行抵抗力[N]</returns>
        public static double RunningResistanceForce(double v)
        {
            return System.Math.Pow(v, 2) * Parameters.Dynamics.RunningResistanceFactorA + v * Parameters.Dynamics.RunningResistanceFactorB + Parameters.Dynamics.RunningResistanceFactorC;
        }

        /// <summary>
        /// ある速度における走行抵抗力による負の加速度[km/h/s]
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public static double RunningResistanceAcceleration(double v)
        {
            return (RunningResistanceForce(v) / OrigMath.TotalInertiaMass) * 3.6;
        }

        /// <summary>
        /// 加速度パラメータ、座標、速度から加速度[km/h/s]の算出
        /// </summary>
        /// <param name="acp">加速度パラメータ</param>
        /// <param name="x">X座標</param>
        /// <param name="width">グラフ領域の横サイズ[px]</param>
        /// <param name="height">グラフ領域の縦サイズ[px]</param>
        /// <param name="xmax">実際のX最大値[km/h]</param>
        /// <param name="ymax">実際のY最大値[km/h/s]</param>
        /// <returns>Y座標</returns>
        public static double CalcAccelerationCurve(Acceleration acp, int x, int width, int height, double xmax, double ymax)
        {
            double v = x / (double)width * xmax; // 速度
            double a; // 加速度

            if (v < acp.ConstPowerStart) // 定引張力域
            {
                a = OrigMath.LinearInterpolation(v, 0.001, acp.ConstAccelerationStart, acp.ConstPowerStart, acp.ConstAccelerationEnd);
            }
            else if (v < acp.ConstPowerEnd) // 定出力域, 0.001はゼロ除算防止
            {
                a = acp.ConstAccelerationEnd * acp.ConstPowerStart / (v + 0.001);
            }
            else // 特性域, 0.001はゼロ除算防止
            {
                a = acp.ConstAccelerationEnd * acp.ConstPowerStart / (acp.ConstPowerEnd + 0.001) * Math.Pow((acp.ConstPowerEnd + 0.001) / (v + 0.001), acp.CharacteristicCurveDegree);
            }
            return (int)(height - (1 + a * height / ymax));
        }

        /// <summary>
        /// 未使用
        /// </summary>
        /// <param name="acp"></param>
        /// <param name="x"></param>
        /// <param name="width"></param>
        /// <param name="height"></param>
        /// <param name="xmax"></param>
        /// <param name="ymax"></param>
        /// <returns></returns>
        public static double CalcAccelerationCurve(Acceleration acp, double x, int width, int height, double xmax, double ymax)
        {
            double v = x / (double)width * xmax; // 速度に変換

            double a = CalcAccelerationCurve(acp, v); // 加速度を算出

            return (int)(height - (1 + a * height / ymax)); // Y座標に変換して戻す
        }

        /// <summary>
        /// 加速度パラメータと速度から加速度[km/h/s]の算出
        /// </summary>
        /// <param name="acp">加速度パラメータ</param>
        /// <param name="v">速度</param>
        /// <returns>加速度</returns>
        public static double CalcAccelerationCurve(Acceleration acp, double v)
        {
            double a; // 加速度

            if (v < acp.ConstPowerStart) // 定引張力域
            {
                a = OrigMath.LinearInterpolation(v, 0.001, acp.ConstAccelerationStart, acp.ConstPowerStart, acp.ConstAccelerationEnd);
            }
            else if (v < acp.ConstPowerEnd) // 定出力域, 0.001はゼロ除算防止
            {
                a = acp.ConstAccelerationEnd * acp.ConstPowerStart / (v + 0.001);
            }
            else // 特性域, 0.001はゼロ除算防止
            {
                a = acp.ConstAccelerationEnd * acp.ConstPowerStart / (acp.ConstPowerEnd + 0.001) * Math.Pow((acp.ConstPowerEnd + 0.001) / (v + 0.001), acp.CharacteristicCurveDegree);
            }
            return a;
        }

        /// <summary>
        /// 加速度パラメータと速度から相対電流値の算出
        /// </summary>
        /// <param name="acp">加速度パラメータ</param>
        /// <param name="v">速度</param>
        /// <returns>相対電流</returns>
        public static double CalcCurrentCurve(Acceleration acp, double v)
        {
            double i; // 電流

            if (v < acp.ConstPowerStart) // 定引張力域
            {
                i = OrigMath.LinearInterpolation(v, 0.001, acp.ConstAccelerationStart, acp.ConstPowerStart, acp.ConstAccelerationEnd);
            }
            else if (v < acp.ConstPowerEnd) // 定出力域
            {
                i = acp.ConstAccelerationEnd;
            }
            else // 特性域, 0.001はゼロ除算防止
            {
                i = acp.ConstAccelerationEnd * Math.Pow((acp.ConstPowerEnd + 0.001) / (v + 0.001), (acp.CharacteristicCurveDegree - 1 + 0.001));
            }
            return i;
        }

        /// <summary>
        /// 加速度[km/h/s]から1両当たり加速力[N]へ変換
        /// </summary>
        /// <param name="a">加速度[km/h/s]</param>
        /// <returns>加速力[N]</returns>
        public static double ForceFromAcceleration(double a)
        {
            return (OrigMath.TotalInertiaMass * (a / 3.6)) / (Parameters.Dynamics.MotorcarCount);
        }

        /// <summary>
        /// 加速度[km/h/s]から満車時の1両当たり加速力[N]へ変換
        /// </summary>
        /// <param name="a"></param>
        /// <returns></returns>
        public static double MaxForceFromAcceleration(double a)
        {
            return ((OrigMath.TotalInertiaMass + (Parameters.Dynamics.MotorcarCount + Parameters.Dynamics.TrailerCount) * Parameters.Original.LoadCompensating) * (a / 3.6)) / (Parameters.Dynamics.MotorcarCount);
        }

        /// <summary>
        /// 1両当たり加速力[N]から加速度[km/h/s]へ変換
        /// </summary>
        /// <param name="f">加速力[N]</param>
        /// <returns>加速度[km/h/s]</returns>
        public static double AccelerationFromForce(double f)
        {
            return (f * (Parameters.Dynamics.MotorcarCount) / OrigMath.TotalInertiaMass) * 3.6;
        }

        /// <summary>
        /// 1両当たり加速力[N]から加速度[km/h/s]へ変換（応荷重込み）
        /// </summary>
        /// <param name="f">加速力[N]</param>
        /// <returns>加速度[km/h/s]</returns>
        public static double MinAccelerationFromForce(double f)
        {
            return (f * (Parameters.Dynamics.MotorcarCount) / OrigMath.MaxTotalInertiaMass) * 3.6;
        }

        /// <summary>
        /// 編成全体の慣性質量[kg]の取得
        /// </summary>
        public static double TotalInertiaMass
        {
            get
            {
                return Parameters.Dynamics.MotorcarCount * Parameters.Dynamics.MotorcarWeight * (Parameters.Dynamics.MotorcarInertiaFactor + 1) + Parameters.Dynamics.TrailerCount * Parameters.Dynamics.TrailerWeight * (Parameters.Dynamics.TrailerInertiaFactor + 1);
            }
        }

        /// <summary>
        /// 応荷重も考慮した編成慣性質量[kg]の取得
        /// </summary>
        public static double MaxTotalInertiaMass
        {
            get
            {
                return TotalInertiaMass + Parameters.Original.LoadCompensating * (Parameters.Dynamics.MotorcarCount + Parameters.Dynamics.TrailerCount);
            }
        }

        // 以下三つはdeligateで簡略化可能？

        /// <summary>
        /// 空車時加速力テーブルの作成
        /// </summary>
        /// <param name="acp"></param>
        /// <param name="interval"></param>
        /// <param name="max"></param>
        /// <returns></returns>
        public static double[,] CreateForceTable(Acceleration[] acp, int interval, int max)
        {
            double[,] table = new double[acp.Length + 1, max / interval + 1];

            for (int i = 0; i < acp.Length + 1; i++)
            {
                /*
                if (acp[i].ConstAccelerationStart == 0 & acp[i].ConstAccelerationEnd == 0)
                {
                    break;
                }
                */
                for (int j = 0; j <= max / interval; j++)
                {
                    if (i == 0)
                    {
                        table[i, j] = j * interval;
                    }
                    else
                    {
                        table[i, j] = OrigMath.ForceFromAcceleration(OrigMath.CalcAccelerationCurve(acp[i - 1], j * interval));
                    }

                }
            }

            return table;
        }

        /// <summary>
        /// 満車時加速力テーブルの作成
        /// </summary>
        /// <param name="acp"></param>
        /// <param name="interval"></param>
        /// <param name="max"></param>
        /// <returns></returns>
        public static double[,] CreateMaxForceTable(Acceleration[] acp, int interval, int max)
        {
            double[,] table = new double[acp.Length + 1, max / interval + 1];

            for (int i = 0; i < acp.Length + 1; i++)
            {
                /*
                if (acp[i].ConstAccelerationStart == 0 & acp[i].ConstAccelerationEnd == 0)
                {
                    break;
                }
                */
                for (int j = 0; j <= max / interval; j++)
                {
                    if (i == 0)
                    {
                        table[i, j] = j * interval;
                    }
                    else
                    {
                        table[i, j] = OrigMath.MaxForceFromAcceleration(OrigMath.CalcAccelerationCurve(acp[i - 1], j * interval));
                    }

                }
            }

            return table;
        }

        /// <summary>
        /// 空車時電流テーブルの作成
        /// </summary>
        /// <param name="acp"></param>
        /// <param name="interval"></param>
        /// <param name="max"></param>
        /// <returns></returns>
        public static double[,] CreateCurrentTable(Acceleration[] acp, int interval, int max, double maxAcceleration)
        {
            double[,] table = new double[acp.Length + 1, max / interval + 1];

            for (int i = 0; i < acp.Length + 1; i++)
            {
                /*
                if (acp[i].ConstAccelerationStart == 0 & acp[i].ConstAccelerationEnd == 0)
                {
                    break;
                }
                */
                for (int j = 0; j <= max / interval; j++)
                {
                    if (i == 0)
                    {
                        table[i, j] = j * interval;
                    }
                    else
                    {
                        table[i, j] = OrigMath.CalcCurrentCurve(acp[i - 1], j * interval) / maxAcceleration * 100;
                    }

                }
            }

            return table;
        }

        /// <summary>
        /// 満車時電流テーブルの作成
        /// </summary>
        /// <param name="current">電流テーブル</param>
        /// <param name="force">加速力テーブル</param>
        /// <param name="maxforce">加速力テーブル（満車時）</param>
        /// <returns></returns>
        public static double[,] CreateMaxCurrentTable(double[,] current, double[,] force, double[,] maxforce)
        {
            double[,] table = new double[current.GetLength(0), current.GetLength(1)];

            for (int i = 0; i < current.GetLength(0); i++)
            {
                for (int j = 0; j < current.GetLength(1); j++)
                {
                    table[i, j] = current[i, j] * maxforce[i, j] / (force[i, j] + 0.001);
                }
            }

            return table;
        }
    }
}
